const mysql = require('mysql2/promise');
const { OpenAI } = require('openai'); 
const express = require('express');
const math = require('mathjs');
const cors = require('cors');
require('dotenv').config();
const fs = require('fs').promises;
const app = express();
app.use(express.json()); 
const port = 3000; 
app.use(cors());


const openai = new OpenAI({ apiKey: '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' });
const dbConfig = {
  host:'localhost',
  user:'aibasedsols_aldenglobal',
  password: 'marketplace@123',
  database: 'aibasedsols_aldenglobal',
};

let Data=[];
async function getData() {
  try {
    const connection = await mysql.createConnection(dbConfig);
    const [rows] = await connection.execute('SELECT * FROM ai_search');
    await connection.end();
    return rows;
  } 
  catch (error) {
    console.error('Database error:', error.message);
    return [];
  }
}

async function preprocessData(rawData) {
    try {
        let structuredData = rawData.map((company) => ({
            companyId: company['Company ID'],
            lead: company.Lead,
            link:company.company_webpage,
            logo:company.Company_logo,
            overview: company.Overview,
            aldenSector: company['Alden Sector'],
            keybenefits: company['Key Benefits'],
            keyFeatures: company['Key Features'],
            challenges: company.Challenges
        }));

        Data = structuredData.filter(row =>
            Object.values(row).every(value => value !== null && value !== undefined)
        );

        Data = Data.map(row => ({
            ...row,
            keywords: `${row.aldenSector} ${row.keyFeatures} ${row.challenges}`.toLowerCase(),
        }));

        const All_Data = 'All_Data.json';
        let savedEmbeddings = {};

        try {
            const fileContent = await fs.readFile(All_Data, 'utf-8');
            savedEmbeddings = JSON.parse(fileContent);
        } catch (error) {
            if (error.code !== 'ENOENT') {
                console.error('Error reading embeddings file:', error.message);
                throw error;
            }
        }

        for (let row of Data) {
            let existingEntry = Object.values(savedEmbeddings).find(savedRow =>
              savedRow.lead === row.lead &&
              savedRow.aldenSector === row.aldenSector &&
              savedRow.keyFeatures === row.keyFeatures &&
              savedRow.challenges === row.challenges
            );
        
            if (existingEntry) {
                row.embedding = existingEntry.embedding;
                row.suggestions = existingEntry.suggestions;
            } else {
                console.log('Row data changed. Computing new embeddings for:', row.lead);
                row.embedding = await getEmbedding(row.keywords);
                let newSuggestions = await getSuggestions(row.keywords);
                row.suggestions = [...new Set([...newSuggestions, ...row.keyFeatures.split(', ')])];

                savedEmbeddings[row.lead] = {
                    lead: row.lead,
                    aldenSector: row.aldenSector,
                    keyFeatures: row.keyFeatures,
                    challenges: row.challenges,
                    suggestions: row.suggestions,
                    embedding: row.embedding
                };
            }
        }

        await fs.writeFile(All_Data, JSON.stringify(savedEmbeddings, null, 2));
        console.log('Preprocessing step is done');
        return Data;
    } catch (error) {
        console.error('Data processing error:', error.message);
    }
}

const cacheFile = "cached_embeddings.json";
async function cacheEmbeddings(text) {
    let cachedData = {};
    try {
        const fileContent = await fs.readFile(cacheFile, 'utf-8');
        cachedData = JSON.parse(fileContent);
    } catch (error) {
        console.error(`Error reading cache file: ${error}. Resetting cache...`);
    }
    if (cachedData[text]) {
        return cachedData[text];
    } else {
        const embedding = await getEmbedding(text);
        cachedData[text] = embedding;
        await fs.writeFile(cacheFile, JSON.stringify(cachedData, null, 2), 'utf-8');
        return embedding;
    }
}

async function getEmbedding(text) {
  try {
    const response = await openai.embeddings.create({
      model: "text-embedding-3-small",
      input: text,
      dimensions:512
    });    
    const embedding = response.data[0].embedding;
    return embedding;
  } catch (error) {
    console.error(`Error fetching embedding: ${error.message}`);
    return Array(512).fill(0);  
  }
}

async function getSuggestions(context, num_suggestions = 15, retries = 3) {
  const prompt = `
    Generate ${num_suggestions} unique autocomplete search queries based on this context: "${context}". 
    Focus on how users naturally enter search queries when looking for specific features, products, or services in this sector. 
    The queries should reflect real user behavior, using concise, intent-driven phrases that include relevant features, benefits, or challenges. 
    Avoid overly generic terms and prioritize phrases that highlight the specific needs or functionality users might search for. 
    Keep the queries short, natural, and directly tied to the context.
  `;
  for (let attempt = 1; attempt <= retries; attempt++) {
    try {
      const response = await openai.chat.completions.create({
        model: "gpt-3.5-turbo",
        messages: [
          { role: "system", content: "You are an assistant generating autocomplete search queries." },
          { role: "user", content: prompt }
        ],
        max_tokens: 200,
        temperature: 0.7,
      });
      const suggestionsText = response.choices[0].message.content.trim();
      const suggestions = suggestionsText
        .split('\n')
        .map(s => s.replace(/^\d+\.\s*/, '').trim())
        .filter(s => s);
      return suggestions.slice(0, 20).map(suggestion => suggestion.replace(/[^a-zA-Z0-9\s]/g, '').trim());
    } catch (error) {
      console.error(`Error generating suggestions for  ${error.message}`);
      if (attempt === retries) {
        return [];
      }
    }
  }
}

async function getUniqueSectors(data) {
  try {
      let uniqueSectors = new Set();   
      data.forEach(row => {
          let sectors = row.aldenSector.split(',');
          sectors.forEach(sector => uniqueSectors.add(sector));
      });
      return Array.from(uniqueSectors).sort();
  } catch (error) {
      console.error(`Error extracting unique sectors: ${error}`);
  }
}

async function main() {
  const rawData = await getData(); 
  if (rawData.length > 0) {
    const Data = await preprocessData(rawData);
    return Data;
  } else {
    console.log('No data found.');
    return [];
  }
}
main();

app.get('/get-sectors', async (req, res) => {
  try {
    const sectors = await getUniqueSectors(Data);
    if (sectors) {
      res.json({ success: true, sectors: sectors });
    } else {
      res.json({ success: false, message: 'No data found.' });
    }
  } catch (error) {
    res.json({ success: false, message: 'Error fetching sectors', error: error.message });
  }
});

app.get('/suggestions', async (req, res) => {
  const suggestionSector = req.query.sector;
  if (!suggestionSector) {
    return res.status(400).json({ error: 'sector not provided' });
  }

  let suggestionsFromCache = [];
  try {
    const cachedData = JSON.parse(await fs.readFile('cached_embeddings.json', 'utf8'));
    suggestionsFromCache = Object.keys(cachedData);
  } catch (error) {
    console.error('Error reading cached embeddings:', error.message);
  }

  try {
    const matchingRows = Data.filter(row =>
      row.aldenSector.split(',').some(sector => sector.toLowerCase() === suggestionSector.toLowerCase())
    );
    const mainsuggestions = matchingRows.flatMap(row => row.suggestions || []);
    const suggestions = Array.from(new Set([...mainsuggestions, ...suggestionsFromCache]));
    return res.json({ suggestions });
  } catch (error) {
    console.error('Error fetching suggestions:', error.message);
    return res.status(500).json({ message: 'Error fetching suggestions', error: error.message });
  }
});

function cosineSimilarity(vec1, vec2) {
  const dotProduct = vec1.reduce((sum, v, i) => sum + v * vec2[i], 0);
  const magnitude1 = Math.sqrt(vec1.reduce((sum, v) => sum + v * v, 0));
  const magnitude2 = Math.sqrt(vec2.reduce((sum, v) => sum + v * v, 0));
  return dotProduct / (magnitude1 * magnitude2);
}

async function get_similarities(sector, inputText) {
  try {
      const similaritysector = sector.toLowerCase();
      const inputEmbedding = await cacheEmbeddings(inputText);
      const matchingRows = Data.filter(row =>
        row.aldenSector && row.aldenSector.split(',').some(s => s.toLowerCase() === similaritysector)
      );
      const similarRows = matchingRows.map(row => ({
          lead: row.lead,
          link:row.link,
          logo:row.logo,
          similarity: cosineSimilarity(inputEmbedding, row.embedding) * 100
      }));
      similarRows.sort((a, b) => b.similarity - a.similarity);
      return similarRows
  } catch (error) {
      console.error("Error computing similar leads:", error);
      return [];
  }
}

async function get_non_filteredLeads(inputText) {
  try {
      const inputEmbedding = await cacheEmbeddings(inputText);
      const similarRows = Data.map(row => ({
          lead: row.lead,
          link: row.link,
          logo: row.logo,
          similarity: cosineSimilarity(inputEmbedding, row.embedding) * 100
      }));
      similarRows.sort((a, b) => b.similarity - a.similarity);
      return similarRows.slice(0, 5);
  } catch (error) {
      console.error("Error computing similar leads:", error);
      return [];
  }
}

app.get('/get_similarities', async (req, res) => {
    const { sector, solutions } = req.query;
    try {
        const similar_Rows = await get_similarities(sector, solutions);
        const filteredLeads = similar_Rows.filter(row => row.similarity > 40);
        
        if (filteredLeads.length > 0) {
            res.json({ success: true, filteredLeads });
        } else {
            const non_filteredLeads = await get_non_filteredLeads(solutions)
            res.json({ success: false, non_filteredLeads });
        }
    } catch (error) {
        console.error(`Error in get_similarities: ${error.message}`);
        res.status(500).json({ error: 'An error occurred while processing the request' });
    }
});

app.listen(port, () => {
  console.log('Serving server on ', {port});
});